import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import math, copy
from torch.autograd import Variable
from torchvision import datasets, models, transforms
from inception import inception_v3
from torch.nn import Parameter

class EncoderDecoder(nn.Module):
    """
    A standard Encoder-Decoder architecture. Base for this and many 
    other models.
    """
    def __init__(self, encoder, decoder, src_embed, tgt_embed, generator):
        super(EncoderDecoder, self).__init__()
        self.encoder = encoder
        self.decoder = decoder
        self.src_embed = src_embed
        self.tgt_embed = tgt_embed
        self.generator = generator
        
    def forward(self, imgs):
        with torch.no_grad():
            imgs = imgs/255.0
            # Comment by Son Trinh 17:17 8/3/2021
            # imgs = imgs.transpose(1, 2).transpose(1, 3).contiguous()
#             src_mask = torch.ones(imgs.shape[0], 1, 377, dtype=torch.bool).cuda()
            src_mask = torch.ones(imgs.size(0), 1, 377).cuda().long()
            ys = torch.ones(imgs.size(0), 1).fill_(1).cuda().long()
            memory = self.encode(imgs, src_mask)
            probs = torch.zeros(imgs.size(0), 1).cuda().float()
            for i in range(11): # 12 -1

                out = self.decode(memory, src_mask, 
                                   ys, 
                                   Variable(subsequent_mask(i + 1).cuda()
                                            .long()))
                logit = self.generator(out[:, -1])
                prob = F.softmax(logit, dim=-1)
                prob_max, next_word = torch.max(prob, dim = 1)
                next_word = next_word.unsqueeze(1)
                ys = torch.cat((ys, next_word), dim=1)
                #ys = torch.from_numpy(ys.cpu().numpy()).cuda().long()
                
                #ys = torch.ones(imgs.size(0), i+1).fill_(1).cuda().long()
                #ys = torch.tensor(ys1).cuda().long()
                probs = torch.cat((probs, prob_max.unsqueeze(1)), dim=1)
        return ys, probs
    
    def encode(self, src, src_mask):
        return self.encoder(self.src_embed(src), src_mask)
    
    def decode(self, memory, src_mask, tgt, tgt_mask):
        return self.decoder(self.tgt_embed(tgt), memory, src_mask, tgt_mask)

class Generator(nn.Module):
    "Define standard linear + softmax generation step."
    def __init__(self, d_model, vocab):
        super(Generator, self).__init__()
        
#         # Arcface head
#         self.weight = Parameter(torch.FloatTensor(vocab, d_model))
#         nn.init.orthogonal_(self.weight)
        
#         self.s = 22
#         self.m = 0.3
#         self.cos_m = math.cos(self.m)
#         self.sin_m = math.sin(self.m)
#         self.mm = math.sin(math.pi-self.m)*self.m
#         self.threshold = math.cos(math.pi-self.m)
#         self.easy_margin = False
        
        self.proj = nn.Linear(d_model, vocab)

    def forward(self, x):
        return self.proj(x)
    
#     def forward(self, x, label):
#         output_class = F.log_softmax(self.proj(x), dim=-1)
#         embedding_norm = F.normalize(x, dim=2)

#         weight_norm = F.normalize(self.weight)
#         zy = F.linear(embedding_norm, weight_norm)*self.s
#         cosine_t = zy/self.s
#         sine_t = torch.sqrt((1.0 - cosine_t*cosine_t).clamp(0 + 1e-6, 1))
#         cos_phi = (cosine_t * self.cos_m - sine_t * self.sin_m)*self.s

#         if self.easy_margin:
#             cos_phi = torch.where(cosine_t > 0, cos_phi, zy)
#         else:
#             cos_phi = torch.where(cosine_t > self.threshold, cos_phi, zy - self.s * self.mm)
        
# #          # --------------------------- convert label to one-hot ---------------------------
#         one_hot = torch.zeros(cosine_t.size(), device='cuda')
       
#         label_new = torch.unsqueeze(label, 2)
# #         # one_hot = one_hot.cuda() if cosine.is_cuda else one_hot

#         one_hot.scatter_(2, label_new, 1)
        
#         output = (one_hot * cos_phi) + ((1 - one_hot) * zy)
#         output_arc = one_hot*(cos_phi-zy) + zy
        
#         return output_arc, output_class


#     def predict(self, x):
#         #print("x: ", x)
#         embedding_norm = F.normalize(x, dim=2)
#         #print("embedding_norm: ", embedding_norm)
#         weight_norm = F.normalize(self.weight)
#         #print("weight norm: ", weight_norm)
#         out = F.linear(embedding_norm, weight_norm)*self.s
#         #print("out shape: ", out)
#         return out

def clones(module, N):
    "Produce N identical layers."
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])

class Encoder(nn.Module):
    "Core encoder is a stack of N layers"
    def __init__(self, layer, N):
        super(Encoder, self).__init__()
        self.layers = clones(layer, N)
        print("Encoder: ", layer.size)
        self.norm = LayerNorm(layer.size)
        
    def forward(self, x, mask):
        "Pass the input (and mask) through each layer in turn."
        for layer in self.layers:
            x = layer(x, mask)
        return self.norm(x)

class LayerNorm(nn.Module):
    "Construct a layernorm module (See citation for details)."
    def __init__(self, features, eps=1e-6):
        super(LayerNorm, self).__init__()
        self.a_2 = nn.Parameter(torch.ones(features))
        self.b_2 = nn.Parameter(torch.zeros(features))
        self.eps = eps

    def forward(self, x):
        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)
        return self.a_2 * (x - mean) / (std + self.eps) + self.b_2

class SublayerConnection(nn.Module):
    """
    A residual connection followed by a layer norm.
    Note for code simplicity the norm is first as opposed to last.
    """
    def __init__(self, size, dropout):
        super(SublayerConnection, self).__init__()
        self.norm = LayerNorm(size)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x, sublayer):
        "Apply residual connection to any sublayer with the same size."
        return x + self.dropout(sublayer(self.norm(x)))

class EncoderLayer(nn.Module):
    "Encoder is made up of self-attn and feed forward (defined below)"
    def __init__(self, size, self_attn, feed_forward, dropout):
        super(EncoderLayer, self).__init__()
        self.self_attn = self_attn
        self.feed_forward = feed_forward
        self.sublayer = clones(SublayerConnection(size, dropout), 2)
        self.size = size

    def forward(self, x, mask):
        "Follow Figure 1 (left) for connections."
        x = self.sublayer[0](x, lambda x: self.self_attn(x, x, x, mask))
        return self.sublayer[1](x, self.feed_forward)

class Decoder(nn.Module):
    "Generic N layer decoder with masking."
    def __init__(self, layer, N):
        super(Decoder, self).__init__()
        self.layers = clones(layer, N)
        print("Decoder: ", layer.size)
        self.norm = LayerNorm(layer.size)
        
    def forward(self, x, memory, src_mask, tgt_mask):
        for layer in self.layers:
            x = layer(x, memory, src_mask, tgt_mask)
        return self.norm(x)

class DecoderLayer(nn.Module):
    "Decoder is made of self-attn, src-attn, and feed forward (defined below)"
    def __init__(self, size, self_attn, src_attn, feed_forward, dropout):
        super(DecoderLayer, self).__init__()
        self.size = size
        self.self_attn = self_attn
        self.src_attn = src_attn
        self.feed_forward = feed_forward
        self.sublayer = clones(SublayerConnection(size, dropout), 3)
 
    def forward(self, x, memory, src_mask, tgt_mask):
        "Follow Figure 1 (right) for connections."
        m = memory
        x = self.sublayer[0](x, lambda x: self.self_attn(x, x, x, tgt_mask))
        x = self.sublayer[1](x, lambda x: self.src_attn(x, m, m, src_mask))
        return self.sublayer[2](x, self.feed_forward)


def subsequent_mask(size):
    "Mask out subsequent positions."
    attn_shape = (1, size, size)
    subsequent_mask = np.triu(np.ones(attn_shape), k=1).astype('uint8')
    return torch.from_numpy(subsequent_mask) == 0

def attention(query, key, value, mask=None, dropout=None):
    "Compute 'Scaled Dot Product Attention'"
    d_k = query.size(-1)
    scores = torch.matmul(query, key.transpose(-2, -1)) \
             / math.sqrt(d_k)
    
    if mask is not None:
        scores = scores.masked_fill(mask == 0, -1e4)
    p_attn = F.softmax(scores, dim = -1)
    if dropout is not None:
        p_attn = dropout(p_attn)
    return torch.matmul(p_attn, value), p_attn

class MultiHeadedAttention(nn.Module):
    def __init__(self, h, d_model, dropout=0.0):
        "Take in model size and number of heads."
        super(MultiHeadedAttention, self).__init__()
        assert d_model % h == 0
        # We assume d_v always equals d_k
        self.d_k = d_model // h
        self.h = h
        self.linears = clones(nn.Linear(d_model, d_model), 4)
        self.attn = None
        self.dropout = nn.Dropout(p=dropout)
        
    def forward(self, query, key, value, mask=None):
        "Implements Figure 2"
        if mask is not None:
            # Same mask applied to all h heads.
            mask = mask.unsqueeze(1)
        nbatches = query.size(0)
        
        # 1) Do all the linear projections in batch from d_model => h x d_k 
        query, key, value = \
            [l(x).view(nbatches, -1, self.h, self.d_k).transpose(1, 2)
             for l, x in zip(self.linears, (query, key, value))]
        
        # 2) Apply attention on all the projected vectors in batch. 
        x, self.attn = attention(query, key, value, mask=mask, 
                                 dropout=self.dropout)
        
        # 3) "Concat" using a view and apply a final linear. 
        x = x.transpose(1, 2).contiguous() \
             .view(nbatches, -1, self.h * self.d_k)
        return self.linears[-1](x)

class PositionwiseFeedForward(nn.Module):
    "Implements FFN equation."
    def __init__(self, d_model, d_ff, dropout=0.1):
        super(PositionwiseFeedForward, self).__init__()
        self.w_1 = nn.Linear(d_model, d_ff)
        self.w_2 = nn.Linear(d_ff, d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        return self.w_2(self.dropout(F.relu(self.w_1(x))))

class Embeddings(nn.Module):
    def __init__(self, d_model, vocab):
        super(Embeddings, self).__init__()
        self.lut = nn.Embedding(vocab, d_model)
        self.d_model = d_model

    def forward(self, x):
        return self.lut(x) * math.sqrt(self.d_model)


class PositionalEncoding(nn.Module):
    "Implement the PE function."
    def __init__(self, d_model, dropout, max_len=377):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)
        # Compute the positional encodings once in log space.
        pe = nn.Parameter(torch.randn(1, max_len, d_model))
        self.register_parameter('pe', pe)
        
    def forward(self, x):
        x = x + self.pe[:, :x.size(1)] 
        return self.dropout(x)
# class PositionalEncoding(nn.Module):
#     def __init__(self, d_model, dropout, max_len=377):
#         super(PositionalEncoding, self).__init__()
#         self.dropout = nn.Dropout(p=dropout)

#         pe = torch.zeros(max_len, d_model)
#         position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
#         div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model))
#         pe[:, 0::2] = torch.sin(position * div_term)
#         pe[:, 1::2] = torch.cos(position * div_term)
#         pe = pe.unsqueeze(0).transpose(0, 1)
#         self.register_buffer('pe', pe)

#     def forward(self, x):
#         x = x + self.pe[:x.size(0), :]

#         return self.dropout(x)


class FeatureExtractor(nn.Module):
    def __init__(self, d_model, d_feature=288):
        super(FeatureExtractor, self).__init__()
        self.conv_base = inception_v3(pretrained=False)
        #self.conv_base = EfficientNet.from_pretrained('efficientnet-b0', in_channels=3)
        self.d_model = d_model
        self.w = nn.Linear(d_feature, d_model)


    def forward(self, x):
        x = self.conv_base(x)
        #print(x.shape)

        b = x.size(0)
        c = x.size(1)
        x = x.view(b, c, -1).permute(0, 2, 1)
        x = self.w(x)
        return x


def make_model(tgt_vocab, N=1, 
               d_model=128, d_ff=256, h=8, dropout=0.1):
    "Helper: Construct a model from hyperparameters."

    c = copy.deepcopy
    attn = MultiHeadedAttention(h, d_model)
    ff = PositionwiseFeedForward(d_model, d_ff, dropout)
    position = PositionalEncoding(d_model, dropout)
    #position = SinusoidalPositionalEmbedding(d_model, 0, init_size=377)
    encoder = Encoder(EncoderLayer(d_model, c(attn), c(ff), dropout), N)
    decoder = Decoder(DecoderLayer(d_model, c(attn), c(attn), 
                             c(ff), dropout), N)
    model = EncoderDecoder(
        encoder, decoder,
        nn.Sequential(FeatureExtractor(d_model), c(position)),
        nn.Sequential(Embeddings(d_model, tgt_vocab), c(position)),
        Generator(d_model, tgt_vocab))
    
    # This was important from their code. 
    # Initialize parameters with Glorot / fan_avg.
    for p in model.parameters():
        if p.dim() > 1:
            nn.init.xavier_uniform(p)
    return model
